""" Module including the class for the segmentation of the images
"""
from typing import List

import numpy as np

from deepness.processing.models.model_base import ModelBase


class Segmentor(ModelBase):
    """Class implements segmentation model

    Segmentation model is used to predict class confidence per pixel of the image.
    """

    def __init__(self, model_file_path: str):
        """

        Parameters
        ----------
        model_file_path : str
            Path to the model file
        """
        super(Segmentor, self).__init__(model_file_path)

    def postprocessing(self, model_output: List) -> np.ndarray:
        """ Postprocess the model output.
        Function returns the mask with the probability of the presence of the class in the image.

        Parameters
        ----------
        model_output : List
            Output from the (Segmentation) model

        Returns
        -------
        np.ndarray
            Batch of postprocessed masks (N,H,W,C), 0-1
        """
        labels = np.clip(model_output[0], 0, 1)

        return labels

    def get_number_of_output_channels(self):
        """ Returns model's number of class

        Returns
        -------
        int
            Number of channels in the output layer
        """
        if len(self.outputs_layers) == 1:
            n_output_channels = self.outputs_layers[0].shape[-3]
            # Support models that return a single output layer, which is converted to 
            # a binary mask.
            if n_output_channels == 1:
                return 2
            else:
                return n_output_channels
        else:
            raise NotImplementedError("Model with multiple output layers is not supported! Use only one output layer.")

    @classmethod
    def get_class_display_name(cls):
        """ Returns the name of the class to be displayed in the GUI

        Returns
        -------
        str
            Name of the class
        """
        return cls.__name__

    def check_loaded_model_outputs(self):
        """ Checks if the model outputs are valid

        Valid means that:
        - the model has only one output
        - the output is 4D (N,C,H,W)
        - the batch size is 1
        - model resolution is equal to TILE_SIZE (is square)

        """
        if len(self.outputs_layers) == 1:
            shape = self.outputs_layers[0].shape

            if len(shape) != 4:
                raise Exception(f'Segmentation model output should have 4 dimensions: (B,C,H,W). Has {shape}')

            if shape[2] != shape[3]:
                raise Exception(f'Segmentation model can handle only square outputs masks. Has: {shape}')

        else:
            raise NotImplementedError("Model with multiple output layers is not supported! Use only one output layer.")
